% Script for comparison of different types of approximation
% of the psi_1 function applied to a matrix.
%
% Generates examples for the manuscript 
% "Computing the Reciprocal of a $\phi$ Function 
% by Rational Approximation",
% by P. Boito, Y. Eidelman and L. Gemignani.
%
% Please note: here accurate computations of psi_1(A) for nonsymmetric A
% for comparison purposes are carried out using the symbolic computation
% toolbox in Matlab (purely symbolic or multiprecision via VPA). Experiments
% in the manuscript use the Advantix toolbox working with 64 digits. 


clear all
close all
% Addpath for expint and chebfun (modify as needed):
% addpath ./expint/common
% addpath /home/paola/Documents/Documents/MATLAB/chebfun-master/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Define psi_1 function:
f=@(z)(z./(exp(z)-1));
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Choose degrees:
np=3;     % polynomial part of mixed approx (# of terms)
ns=50;    % rational part of mixed approx (# of terms)
k=7;      % degree of diagonal Padé approx



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Define test matrix a and compute b=psi_1(a).
% The algorithm to be used for computation of b 
% depends on the choice of a.
% Warning: for some matrices it is not easy to
% compute b in a reliable way!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% % Example from Al-Mohy and Higham, 2009:
% n=2;
% beta=1e+5;
% a=[1 beta; 0 -1];

% % Other example
% n=50;
% n2=25;
% a=[eye(n2),1e+4*ones(n2);zeros(n2),-eye(n2)];

% % Random matrix:
% n=50;
% a=-rand(n);
% aexp=expm(a);
% b=(aexp-eye(n))\a;

% % Hanowa matrix from gallery:
% n=50;
% a=gallery('hanowa',n);
% aexp=expm(a);
% b=(aexp-eye(n))\a;

 % Poisson matrix from gallery:
 n=900;
 a=full(gallery('poisson',30));
 [v,d]=eig(a);
 max(abs(diag(d)))
 b=v*diag(f(diag(d)))*v';

%  % Kac-Murdock-Szego matrix from gallery:
%   n=256;
%   a=0.5*full(gallery('kms',n,0.8));
%   [v,d]=eig(a);
%   b=v*diag(f(diag(d)))*v';
  

% % Tridiagonal matrix from gallery:
%  n=256;
%  a=full(gallery('tridiag',n,-1,4,-1));
%  [v,d]=eig(a);
% % max(abs(diag(d)))
%  b=v*diag(f(diag(d)))*v';
 
 % % Inverse of tridiagonal matrix from gallery:
%  n=256;
%  a = 0.7 *inv(gallery('tridiag',n, (n/2)*ones(n-1,1),[n:-1:1], (n/2)*ones(n-1,1)));
%  a=full(a);
%  [v,d]=eig(a);
%  % max(abs(diag(d)))
%  b=v*diag(f(diag(d)))*v';
 

% % Other nonsymmetric matrix from gallery:
% n=100;
% a=full(gallery('smoke',n));


% Scaled companion matrices:
% n=1024;
% gamma=64;
% a= gamma*compan([1; zeros(n-1,1); -1]); 
% [v,d]=eig(a);
% b=v*diag(f(diag(d)))*v';


% Accurate computation of b=psi_1(a):
% as=sym(a);
% aexp=funm(as,@exp);
% bs=as/(aexp-eye(size(a)));

%  nd=32;
%  as=vpa(a,nd);
%  aexp=funm(as,@exp);
%  bs=as/(aexp-vpa(eye(size(a)),nd));
% %%b=funm(a,f);
% %b=as/(aexp-eye(n));
% %bphi=(aexp-eye(n))/as;
% %bphi=double(bphi);
% %b=double(bphi);
% b=double(bs);
% plot(eig(a),'r*');

% Compute polynomial and mixed approximations.
[bmix,bpoly]=mixed_approx(a, np, ns);
bmixss=mixed_approx_ws(a, np, ns);
% Compute errors:
disp('Relative error for polynomial approximation:');
disp(norm(bpoly-b)/norm(b));
disp('Absolute error for polynomial approximation:');
disp(norm(bpoly-b));
disp('Relative error for mixed approximation:');
disp(norm(bmix-b)/norm(b));
disp('Absolute error for mixed approximation:');
disp(norm(bmix-b));
disp('Relative error for mixed approximation with scaling:');
disp(norm(bmixss-b)/norm(b));
disp('Absolute error for mixed approximation with scaling:');
disp(norm(bmixss-b));

% Padé approx using EXPINT:
bpade=phipade(a,1,k);
bpadeinv=inv(bpade);
disp('Padé approx using EXPINT.')
disp('Relative error on approx of psi_1(a):')
disp(norm(bpadeinv-b)/norm(b));
disp('Absolute error on approx of psi_1(a):')
disp(norm(bpadeinv-b));
disp('Condition number of approx of phi_1(a):')
disp(cond(bpade));
%disp('Relative error on approx of phi_1(a):')
%disp(norm(bpade-bphi)/norm(bphi));
%disp('Absolute error on approx of phi_1(a):')
%disp(norm(bpade-bphi));

% Padé approx using EXPM:
bpadeexpm=a/(expm(a)-eye(size(a)));
disp('Padé approx using EXPM.')
disp('Relative error on approx of psi_1(a):')
disp(norm(bpadeexpm-b)/norm(b));
disp('Absolute error on approx of psi_1(a):')
disp(norm(bpadeexpm-b));
